#INCLUDE "dbc.h"

*!*	<pdm_sc_yes/>
*!*	<pdm_dd_yes/>
******************************************************************
**
** DBC low level function
** 
** Version: 0.1.0
** 
** Author: Martin Jindra
** E-mail: martin.jindra@egservis.cz
** Date:   2005-08-08
** 
** Acknowledgments:
** Doug Henning: SPECIAL REPORT - Visual FoxPro Data Dictionary,
**               Pinnacle Publishing Inc.,
**               Kent,1995, ISBN 1-880935-41-4 
**
******************************************************************


*********************************************************************************
*
*********************************************************************************
DEFINE CLASS _DBC AS SESSION && Base dbc class
   Name="_DBC" && Name of class

   PROTECTED _DataSessionID
   _DataSessionID=0 && Backup datasession ID 
   DataSession=2 && Private data session
   Catch=.NULL. && Catch flag
   nErr=0 && Error number
   cErrMSG="" && Error message
   oConnector=.NULL. && Object reference to connector which can  (support for user custom properties)...
  
   PROCEDURE Init() && Backup session ID
      *!*   <pdm_sc_yes/>
      This._DataSessionID=This.DataSessionID
   ENDPROC

   PROCEDURE Destroy() && Restore Datasession ID
      *!*   <pdm_sc_yes/>
      IF This.DataSessionID#This._DataSessionID
         This.DataSessionID=This._DataSessionID
      ENDIF
   ENDPROC


   PROCEDURE GetIDObject(lcAlias,lcType,lcObjName) && Get Object ID
      * lcAlias   - Alias of opened DBC
      * lcType    - Object Type
      * lcObjName - Object Name

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      SELE (lcAlias) && Skip to DBC 
      IF INLIST(lcType,_DBC_Field,_DBC_Index,_DBC_Relation)
         LOCAL lcParent,lii,lcObj
         lii=AT(".",lcObjName) && Find dot
         lcParent=PADR(UPPER(LEFT(lcObjName,lii-1)),128)
         lcObj=PADR(UPPER(SUBST(lcObjName,lii+1)),128)
         * Find Parent
         LOCATE FOR INLIST(ObjectType,_DBC_Table,_DBC_View) AND;
                    UPPER(ObjectName)==lcParent  AND ! DELE()
         IF FOUND() && If found
            lii=ObjectID
            * Find object
            LOCATE FOR ParentID=lii AND ObjectType==lcType AND ;
                       UPPER(ObjectName)==lcObj AND ! DELE()
         ENDIF
      ELSE
         LOCATE FOR ObjectType==lcType AND UPPER(ObjectName)==PADR(UPPER(lcObjName),128);
                AND ! DELE()
      ENDIF
      RETURN IIF(FOUND(),ObjectID,_DBC_ObjectNotFound)
   ENDPROC


   PROCEDURE EnumObjects() && Get list of object by type
      LPARAMETERS lcAlias,lcType,liIDParent,laObjects
      * lcAlias    - Alias of opened DBC
      * lcType     - Object Type
      * liIDParent - Parent ID
      * @laObjects - Output array for object

      EXTERNAL ARRAY laObjects

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      IF EMPTY(lcType)
         RETURN _DBC_ErrParamFailed
      ENDIF
      LOCAL llFound,liCount,lii,liID
      liCount=0
      SELE (lcAlias) && Skip to DBC 
      IF INLIST(lcType,_DBC_Field,_DBC_Index,_DBC_Relation)
         SET FILTER TO ParentID=liIDParent AND ObjectType==lcType AND ! DELE()
      ELSE
         SET FILTER TO ObjectType==lcType AND ! DELE()
      ENDIF
      GO TOP
      COUNT TO liCount
      IF liCount>0
         DIME laObjects(liCount,2)
         GO TOP
         lii=0
         SCAN ALL
              lii=lii+1
              laObjects(lii,1)=OBJECTID
              laObjects(lii,2)=ALLT(OBJECTNAME)
         ENDSCAN
      ENDIF
      SET FILTER TO
      RETURN liCount
   ENDPROC


   PROCEDURE EnumProperty() && Get list of object properties
      LPARAMETERS lcAlias,liIDObject,laProp
      * lcAlias    - Alia sof opened DBC
      * liIDObject - Object ID
      * @laProp    - Output array for Properties and their values

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      EXTERNAL ARRAY laProp

      LOCAL liStart,liLen,liLenID,lii
      liStart=1 && First char

      SELE (lcAlias) && Skip to DBC 
      LOCATE FOR ObjectID=liIDObject && Find object
      IF !FOUND() && If not found
         RETURN _DBC_ObjectNotFound && Get out
      ENDIF
      lii=0 && Reset counter
      DO WHILE liStart<=LEN(Property)
         liLen=This.C4ToI(SUBS(Property,liStart,4)) && Lenght
         * This is lenght of Property ID
         liLenID=IIF(This.C2ToI(SUBS(Property,liStart+4,2))>256,2,1)

         lii=lii+1 && Increment counter
         DIME laProp(lii,3)
         * Property ID
         laProp(lii,1)=IIF(liLenID=1,ASC(SUBS(Property,liStart+4+2,liLenID)),;
                       This.C2ToI(SUBS(Property,liStart+4+2,liLenID)))
         * Convert DBC value to Value
         laProp(lii,2)=This.DBCValueToValue(laProp(lii,1),;
                       SUBS(Property,liStart+3+2+liLenID+1,liLen-(3+2+liLenID+1)))
         liStart=liStart+liLen
      ENDDO
      RETURN lii
   ENDPROC


   PROCEDURE RIInfo() && Get RI Info for Relation
      LPARAMETERS lcAlias,liIDObject,luValue
      * lcAlias    - Alias of opened DBC
      * liIDObject - Object ID
      * @luValue   - (output) RI Info

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      SELE (lcAlias) && Skip to DBC 
      LOCATE FOR ObjectID=liIDObject && Find object
      IF !FOUND() && If not found
         RETURN _DBC_ObjectNotFound && Get out
      ENDIF
      IF OBJECTTYPE#_DBC_Relation
         RETURN _DBC_ObjectNotType
      ENDIF
      luValue=RIINFO
      RETURN _DBC_ObjectFound
   ENDPROC


   PROCEDURE GetSP(lcAlias,liType,luValue) && Get Stored Procedure
      * lcAlias  - Alias of opened DBC
      * liType   - Type (Source code/ P-Code)
      * @luValue - (output) Stored procedures source code/P-Code

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      SELE (lcAlias) && Skip to DBC 
      * Find object
      LOCATE FOR OBJECTTYPE=_DBC_DataBase AND;
             UPPER(ObjectName)==UPPER(IIF(liType=_DBC_SP_TSource,_DBC_SP_Source,_DBC_SP_FXP))
      IF !FOUND() && If not found
         RETURN _DBC_ObjectNotFound && Get out
      ENDIF
      luValue=CODE
      RETURN _DBC_ObjectFound
   ENDPROC


   PROCEDURE SetSP(lcAlias,liType,luValue) && Set Stored Procedure
      * lcAlias  - Alias of opened DBC
      * liType   - Type (Source code/ P-Code)
      * @luValue - (output) Stored procedures source code/P-Code

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      SELE (lcAlias) && Skip to DBC 
      * Find object
      LOCATE FOR OBJECTTYPE=_DBC_DataBase AND ;
             UPPER(ObjectName)==UPPER(IIF(liType=_DBC_SP_TSource,_DBC_SP_Source,_DBC_SP_FXP))
      IF !FOUND() && If not found
         RETURN _DBC_ObjectNotFound && Get out
      ENDIF
      REPLACE CODE WITH luValue
      RETURN _DBC_ObjectFound
   ENDPROC


   PROCEDURE GetProperty() && Get Property's value
      LPARAMETERS lcAlias,liIDObject,liIDProp,luValue
      * lcAlias    - Alias of opened DBC
      * liIDObject - Object ID
      * liIDProp   - Property ID
      * @luValue   - (output) Value of property

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      LOCAL liStart,liLen,liLenID
      liStart=1 && First char

      SELE (lcAlias) && Skip to DBC 
      LOCATE FOR ObjectID=liIDObject && Find object
      IF !FOUND() && If not found
         RETURN _DBC_ObjectNotFound && Get out
      ENDIF
      DO WHILE liStart<=LEN(Property)
         liLen=This.C4ToI(SUBS(Property,liStart,4)) && Lenght
         * This is lenght of Property ID
         liLenID=This.C2ToI(SUBS(Property,liStart+4,2))

         * Is't finding property
         IF This.CnToI(SUBS(Property,liStart+4+2,liLenID))=liIDProp
            luValue=This.DBCValueToValue(liIDProp,;
                    SUBS(Property,liStart+4+2+liLenID,liLen-(4+2+liLenID)) )
            RETURN _DBC_ObjectFound
         ENDIF
         liStart=liStart+liLen
      ENDDO
      luValue=This.GetDefaultValue(liIDProp)
      RETURN _DBC_ObjectFound
   ENDPROC


   PROCEDURE SetProperty(lcAlias,liIDObject,liIDProp,luValue) && Set Value of property
      * lcAlias    - Alias of opened DBC
      * liIDObject - Object ID
      * liIDProp   - Property ID
      * luValue    -  Value

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      LOCAL liStart,liLen,liLenID,lcPom,liType
      liStart=1 && First char

      SELE (lcAlias) && Skip to DBC 
      LOCATE FOR ObjectID=liIDObject && Find object
      IF !FOUND() && If not found
         RETURN _DBC_ObjectNotFound && Get out
      ENDIF

      * Convert value to DBC value
      luValue=This.ValueToDBCValue(liIDProp,luValue)

      IF ISNULL(luValue)
         RETURN _DBC_ValueInvalid
      ENDIF

      DO WHILE liStart<=LEN(Property)
         liLen=This.C4ToI(SUBS(Property,liStart,4)) && Lenght
         * This is lenght of Property ID
         liLenID=This.C2ToI(SUBS(Property,liStart+4,2))

         * Is't finding property
         IF This.CnToI(SUBS(Property,liStart+4+2,liLenID))=liIDProp
            REPL Property WITH STUFF(Property,liStart,liLen,;
                          This.IToC4(4+2+liLenID+LEN(luValue))+;
                          This.IToC2(liLenID)+;
                          This.IToCn(liIDProp)+luValue)
            RETURN _DBC_ObjectFound
         ENDIF
         liStart=liStart+liLen
      ENDDO
      lcPom=This.IToCn(liIDProp)
      REPL Property WITH Property+;
                         This.IToC4(4+2+LEN(lcPom)+LEN(luValue))+;
                         This.IToC2(LEN(lcPom))+;
                         lcPom+luValue
      RETURN _DBC_ObjectFound
   ENDPROC


   PROCEDURE ClearProperty(lcAlias,liIDObject,liIDProp) && Clear property
      * lcAlias    - Alias of opened DBC
      * liIDObject - Object ID
      * liIDProp   - Property ID

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      LOCAL liStart,liLen,liLenID
      liStart=1 && First char

      SELE (lcAlias) && Skip to DBC 
      LOCATE FOR ObjectID=liIDObject && Find object
      IF !FOUND() && If not found
         RETURN _DBC_ObjectNotFound && Get out
      ENDIF

      DO WHILE liStart<=LEN(Property)
         liLen=This.C4ToI(SUBS(Property,liStart,4)) && Lenght
         * This is lenght of Property ID
         liLenID=This.C2ToI(SUBS(Property,liStart+4,2))

         * Is't finding property
         IF This.CnToI(SUBS(Property,liStart+4+2,liLenID))=liIDProp
            REPL Property WITH STUFF(Property,liStart,liLen,"")
            EXIT
         ENDIF
         liStart=liStart+liLen
      ENDDO

      RETURN _DBC_ObjectFound
   ENDPROC


   PROCEDURE CheckProperties() && Check Property's value
      LPARAMETERS lcAlias,liIDObject,laProp
      * lcAlias    - Alias of opened DBC
      * liIDObject - Object ID
      * @laProp    - Output array for Properties, their values and results

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      LOCAL liStart,liLen,liLenID,lii,liID,liType
      liStart=1 && First char

      SELE (lcAlias) && Skip to DBC 
      LOCATE FOR ObjectID=liIDObject && Find object
      IF !FOUND() && If not found
         RETURN _DBC_ObjectNotFound && Get out
      ENDIF
      lii=0 && Reset counter
      DO WHILE liStart<=LEN(Property)
         liLen=This.C4ToI(SUBS(Property,liStart,4)) && Lenght
         * This is lenght of Property ID
         liLenID=This.C2ToI(SUBS(Property,liStart+4,2))
         
         lii=lii+1
         DIMENSION laProp(lii,4)
         * Property ID
         laProp(lii,1)=This.CnToI(SUBS(Property,liStart+4+2,liLenID))

         * Native Value
         laProp(lii,3)=SUBS(Property,liStart+3+2+liLenID+1,liLen-(3+2+liLenID+1))

         * Check ID property
         liType=This.GetValueType(laProp(lii,1))
         IF liType=_DBC_ValueType_Unknown
            laProp(lii,4)=_DBC_Check_PropertyType_Unknown
         ELSE
            * Check if native value have right format
            IF IIF(liType=_DBC_ValueType_String, RIGHT(laProp(lii,3),1)#CHR(0),;
               IIF(liType=_DBC_ValueType_Byte, LEN(laProp(lii,3))#1,;
               IIF(liType=_DBC_ValueType_SI32R,LEN(laProp(lii,3))#4,;
               IIF(liType=_DBC_ValueType_I32R, LEN(laProp(lii,3))#4,;
               IIF(liType=_DBC_ValueType_Boolean, ;
                   LEN(laProp(lii,3))#1 OR !INLIST(ASC(laProp(lii,3)),0,1),;
               IIF(!ISNULL(This.oConnector),This.oConnector.CheckProperties(laProp(lii,1),liType,laProp(lii,3)),;
               .T.))))))
               laProp(lii,4)=_DBC_Check_PropertyValue_Invalid
            ELSE
               * Convert DBC value to Value
               laProp(lii,2)=This.DBCValueToValue(laProp(lii,1),laProp(lii,3))
               laProp(lii,4)=_DBC_Check_OK
            ENDIF
         ENDIF
         liStart=liStart+liLen
      ENDDO
      RETURN lii
   ENDPROC



   PROCEDURE IsBaseType(liType) && Return is value type is base or not
      * liType     - Value type

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>
       
      RETURN INLIST(liType,;
          _DBC_ValueType_Unknown,;
          _DBC_ValueType_String,;
          _DBC_ValueType_Byte,;
          _DBC_ValueType_SI32R,;
          _DBC_ValueType_I32R,;
          _DBC_ValueType_Boolean)
   ENDPROC

   PROCEDURE GetValueType(liIDProp) && Return the property type
      * liIDProp - Property ID

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      LOCAL liRet
      liRet=IIF(INLIST(liIDProp,_DBCID_Comment,_DBCID_Path,;
                _DBCID_RuleExpression,_DBCID_RuleText,_DBCID_InsertTrigger,;
                _DBCID_UpdateTrigger,_DBCID_DeleteTrigger,_DBCID_PrimaryKey,;
                _DBCID_DefaultValue,_DBCID_Caption,_DBCID_ChildTag,;
                _DBCID_ParentTable,_DBCID_Database) OR ;
                INLIST(liIDProp,_DBCID_ParentTag,_DBCID_DataSource,;
       	       _DBCID_ConnectString,_DBCID_Password,_DBCID_UserID,;
         	    _DBCID_ParameterList,_DBCID_ConnectName,_DBCID_SQL,;
                _DBCID_Tables,_DBCID_UpdateName,_DBCID_DataType,;
                _DBCID_DisplayClass,_DBCID_DisplayClassLibrary,;
                _DBCID_Format,_DBCID_InputMask,;
                _DBCID_DBCEventFileName),_DBC_ValueType_String,;
            IIF(INLIST(liIDProp,_DBCID_SubType,_DBCID_DispLogin,;
                _DBCID_Transactions,_DBCID_SourceType,;
                _DBCID_UpdateType,_DBCID_WhereType),_DBC_ValueType_Byte,;
            IIF(INLIST(liIDProp,_DBCID_FetchSize,_DBCID_MaxRecords),_DBC_ValueType_SI32R,;
            IIF(INLIST(liIDProp,_DBCID_ConnectTimeout,_DBCID_IdleTimeout,_DBCID_QueryTimeout,;
                _DBCID_WaitTime,_DBCID_PacketSize,_DBCID_UseMemoSize,;
                _DBCID_Version,_DBCID_BatchUpdateCount),_DBC_ValueType_I32R,;
            IIF(INLIST(liIDProp,_DBCID_Prim_Cand,_DBCID_Asynchronous,_DBCID_BatchMode,;
                _DBCID_DispWarnings,_DBCID_FetchMemo,_DBCID_ShareConnection,;
                _DBCID_SendUpdates,_DBCID_KeyField,_DBCID_Updatable,;
                _DBCID_CompareMemo,_DBCID_FetchAsNeeded,_DBCID_Offline,;
                _DBCID_Prepared,_DBCID_DBCEvents,_DBCID_AllowSimultaneousFetch,;
                _DBCID_DisconnectRollback),_DBC_ValueType_Boolean,_DBC_ValueType_Unknown)))))

      RETURN IIF(liRet=_DBC_ValueType_Unknown AND !ISNULL(This.oConnector),;
                 This.oConnector.GetValueType(liIDProp),liRet)
   ENDPROC


   PROCEDURE DBCValueToValue(liIDProp,lcValue) && Convert DBC value to Value
      * liIDProp - Property ID
      * luValue  - Property value

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      LOCAL liType
      liType=This.GetValueType(liIDProp)

      IF !This.IsBaseType(liType) AND !ISNULL(This.oConnector)
         RETURN This.oConnector.DBCValueToValue(liType,@lcValue)
      ENDIF

      RETURN IIF(liType=_DBC_ValueType_String,LEFT(lcValue,LEN(lcValue)-1),;
             IIF(liType=_DBC_ValueType_Byte,ASC(lcValue),;
             IIF(liType=_DBC_ValueType_SI32R,;
                 IIF(lcValue=CHR(255)+CHR(255)+CHR(255)+CHR(255),-1,This.C4ToIRev(@lcValue)),;
             IIF(liType=_DBC_ValueType_I32R,This.C4ToIRev(@lcValue),;
             IIF(liType=_DBC_ValueType_Boolean,ASC(lcValue)=1,lcValue)))))

   ENDPROC


   PROCEDURE ValueToDBCValue(liIDProp,luValue) && Convert value to DBC value
      * liIDProp - Property ID
      * luValue  - Property value

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>


      LOCAL liType
      liType=This.GetValueType(liIDProp)

      * Check value if is valid
      IF !This.CheckValue(liIDProp,liType,luValue)
         RETURN .NULL.
      ENDIF

      IF !This.IsBaseType(liType) AND !ISNULL(This.oConnector)
         RETURN This.oConnector.ValueToDBCValue(liType,luValue)
      ENDIF

      RETURN IIF(ISNULL(luValue),.NULL.,;
             IIF(liType=_DBC_ValueType_String ,luValue+CHR(0),;
             IIF(liType=_DBC_ValueType_Byte,CHR(luValue),;
             IIF(liType=_DBC_ValueType_SI32R,;
                 IIF(luValue=-1,CHR(255)+CHR(255)+CHR(255)+CHR(255),This.IToC4Rev(@luValue)),;
             IIF(liType=_DBC_ValueType_I32R,This.IToC4Rev(@luValue),;
             IIF(liType=_DBC_ValueType_Boolean,IIF(luValue,CHR(1),CHR(0)),.NULL.))))))
   ENDPROC



   PROCEDURE CheckValue() && Check Value
      LPARAMETERS liIDProp,liType,luValue
      * liIDProp   - Property ID
      * liType     - Value type
      * luValue    - Value

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>
 
      * 1) check type
      * 2) interval
      * 3) if type is not base, call connector if defined
      * 4) return .T. is value valid or .F. is invalid
       
      IF (liType=_DBC_ValueType_String AND (!INLIST(TYPE("luValue"),"C","M") OR ISNULL(luValue))) OR ;
         (liType=_DBC_ValueType_Byte AND (!TYPE("luValue")="N" OR ISNULL(luValue))) OR ;
         (liType=_DBC_ValueType_SI32R AND (!TYPE("luValue")="N" OR ISNULL(luValue))) OR ;
         (liType=_DBC_ValueType_I32R AND (!TYPE("luValue")="N" OR ISNULL(luValue))) OR ;
         (liType=_DBC_ValueType_Boolean AND (!TYPE("luValue")="L" OR ISNULL(luValue)))
         RETURN .F.
      ENDIF

      DO CASE
         CASE INLIST(liIDProp,_DBCID_Comment,_DBCID_Path,;
              _DBCID_RuleExpression,_DBCID_RuleText,_DBCID_InsertTrigger,;
              _DBCID_UpdateTrigger,_DBCID_DeleteTrigger,_DBCID_PrimaryKey,;
              _DBCID_DefaultValue,_DBCID_Caption,_DBCID_ChildTag,;
              _DBCID_ParentTable,_DBCID_Database) OR ;
              INLIST(liIDProp,_DBCID_ParentTag,_DBCID_DataSource,;
              _DBCID_ConnectString,_DBCID_Password,_DBCID_UserID,;
              _DBCID_ParameterList,_DBCID_ConnectName,_DBCID_SQL,;
              _DBCID_Tables,_DBCID_UpdateName,_DBCID_DataType,;
              _DBCID_DisplayClass,_DBCID_DisplayClassLibrary,;
              _DBCID_Format,_DBCID_InputMask,;
              _DBCID_DBCEventFileName)
              RETURN .T.

         CASE liIDProp=_DBCID_SubType
              RETURN INLIST(luValue,1,6,7)

         CASE liIDProp=_DBCID_DispLogin
              RETURN INLIST(luValue,1,2,3)

         CASE liIDProp=_DBCID_Transactions
              RETURN INLIST(luValue,1,2)

         CASE liIDProp=_DBCID_SourceType
              RETURN INLIST(luValue,1,2)

         CASE liIDProp=_DBCID_UpdateType
              RETURN INLIST(luValue,1,2)

         CASE liIDProp=_DBCID_WhereType
              RETURN INLIST(luValue,1,2,3,4)

         CASE liIDProp=_DBCID_FetchSize
              RETURN luValue=-1 OR BETWEEN(luValue,0,0xFFFFFFF)

         CASE liIDProp=_DBCID_MaxRecords
              RETURN luValue=-1 OR BETWEEN(luValue,0,0xFFFFFFF)

         CASE liIDProp=_DBCID_ConnectTimeout
              RETURN BETWEEN(luValue,0,600)

         CASE liIDProp=_DBCID_IdleTimeout
              RETURN BETWEEN(luValue,0,0xA8C1)

         CASE liIDProp=_DBCID_QueryTimeout
              RETURN BETWEEN(luValue,0,0x600)

         CASE liIDProp=_DBCID_WaitTime
              RETURN BETWEEN(luValue,0,0x7FFFFFFF)

         CASE liIDProp=_DBCID_PacketSize
              RETURN BETWEEN(luValue,0,0x7FFFFFFF)

         CASE liIDProp=_DBCID_UseMemoSize
              RETURN BETWEEN(luValue,1,255)

         CASE liIDProp=_DBCID_Version
              RETURN INLIST(luValue,10,11)

         CASE liIDProp=_DBCID_BatchUpdateCount
              RETURN BETWEEN(luValue,0,0xFFFFFFF)

         CASE INLIST(liIDProp,_DBCID_Prim_Cand,_DBCID_Asynchronous,_DBCID_BatchMode,;
              _DBCID_DispWarnings,_DBCID_FetchMemo,_DBCID_ShareConnection,;
              _DBCID_SendUpdates,_DBCID_KeyField,_DBCID_Updatable,;
              _DBCID_CompareMemo,_DBCID_FetchAsNeeded,_DBCID_Offline,;
              _DBCID_Prepared,_DBCID_DBCEvents,_DBCID_AllowSimultaneousFetch,;
              _DBCID_DisconnectRollback)
              RETURN .T.

         OTHERWISE !ISNULL(This.oConnector)
              RETURN This.oConnector.CheckValue(liIDProp,liType,luValue)

      ENDCASE
      RETURN .F.
   ENDPROC

   PROCEDURE GetDefaultValue(liIDProp) && Get default Value
      * liIDProp   - Property ID

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>
      LOCAL liRet
      liRet=IIF(INLIST(liIDProp,_DBCID_Comment,_DBCID_Path,;
                _DBCID_RuleExpression,_DBCID_RuleText,_DBCID_InsertTrigger,;
                _DBCID_UpdateTrigger,_DBCID_DeleteTrigger,_DBCID_PrimaryKey,;
                _DBCID_DefaultValue,_DBCID_Caption,_DBCID_ChildTag,;
                _DBCID_ParentTable,_DBCID_Database) OR ;
                INLIST(liIDProp,_DBCID_ParentTag,_DBCID_DataSource,;
       	       _DBCID_ConnectString,_DBCID_Password,_DBCID_UserID,;
       	       _DBCID_ParameterList,_DBCID_ConnectName,_DBCID_SQL,;
                _DBCID_Tables,_DBCID_UpdateName,_DBCID_DataType,;
                _DBCID_DisplayClass,_DBCID_DisplayClassLibrary,;
                _DBCID_Format,_DBCID_InputMask,_DBCID_DBCEventFileName),"",;
            IIF(INLIST(liIDProp,_DBCID_SubType,_DBCID_ConnectTimeout,;
                _DBCID_IdleTimeout,_DBCID_QueryTimeout),0,;
            IIF(INLIST(liIDProp,_DBCID_DispLogin,_DBCID_Transactions,;
                _DBCID_BatchUpdateCount,;
                _DBCID_SourceType,_DBCID_UpdateType,_DBCID_WhereType),1,;
            IIF(liIDProp=_DBCID_MaxRecords,-1,;
            IIF(liIDProp=_DBCID_Version,10,;
            IIF(INLIST(liIDProp,_DBCID_WaitTime,_DBCID_FetchSize),100,;
            IIF(liIDProp=_DBCID_PacketSize,4096,;
            IIF(liIDProp=_DBCID_UseMemoSize,255,;
            IIF(INLIST(liIDProp,_DBCID_DispWarnings,_DBCID_ShareConnection,;
                _DBCID_CompareMemo),.T.,;
            IIF(INLIST(liIDProp,_DBCID_Prim_Cand,_DBCID_Asynchronous,_DBCID_BatchMode,;
                _DBCID_FetchMemo,_DBCID_AllowSimultaneousFetch,_DBCID_DisconnectRollback,;
                _DBCID_SendUpdates,_DBCID_KeyField,_DBCID_Updatable,;
                _DBCID_FetchAsNeeded,_DBCID_Offline,;
                _DBCID_Prepared,_DBCID_DBCEvents),.F.,.NULL.))))))))))

      RETURN IIF(ISNULL(liRet) AND !ISNULL(This.oConnector),;
                 This.oConnector.GetDefaultValue(liIDProp),liRet)
   ENDPROC


   PROCEDURE FormatProperty(liIDProp) && Convert property ID to Name
      * liIDProp   - Property ID

      *!*	<pdm_sc_yes/>
      *!*	<pdm_dd_yes/>

      DO CASE
         CASE liIDProp=_DBCID_Path
              RETURN "Path"

         CASE liIDProp=_DBCID_SubType
              RETURN "SubType"

         CASE liIDProp=_DBCID_Comment
              RETURN "Comment"

         CASE liIDProp=_DBCID_RuleExpression
              RETURN "RuleExpression"

         CASE liIDProp=_DBCID_RuleText
              RETURN "RuleText"

         CASE liIDProp=_DBCID_DefaultValue
              RETURN "DefaultValue"

         CASE liIDProp=_DBCID_ParameterList
              RETURN "ParameterList"

         CASE liIDProp=_DBCID_ChildTag
              RETURN "ChildTag"

         CASE liIDProp=_DBCID_InsertTrigger
              RETURN "InsertTrigger"

         CASE liIDProp=_DBCID_UpdateTrigger
              RETURN "UpdateTrigger"

         CASE liIDProp=_DBCID_DeleteTrigger
              RETURN "DeleteTrigger"

         CASE liIDProp=_DBCID_Prim_Cand
              RETURN "Prim_Cand"

         CASE liIDProp=_DBCID_ParentTable
              RETURN "ParentTable"

         CASE liIDProp=_DBCID_ParentTag
              RETURN "ParentTag"

         CASE liIDProp=_DBCID_PrimaryKey
              RETURN "PrimaryKey"

         CASE liIDProp=_DBCID_Version
              RETURN "Version"

         CASE liIDProp=_DBCID_BatchUpdateCount
              RETURN "BatchUpdateCount"

         CASE liIDProp=_DBCID_DataSource
              RETURN "DataSource"

         CASE liIDProp=_DBCID_ConnectName
              RETURN "ConnectName"

         CASE liIDProp=_DBCID_UpdateName
              RETURN "UpdateName"

         CASE liIDProp=_DBCID_FetchMemo
              RETURN "FetchMemo"

         CASE liIDProp=_DBCID_FetchSize
              RETURN "FetchSize"

         CASE liIDProp=_DBCID_KeyField
              RETURN "KeyField"

         CASE liIDProp=_DBCID_MaxRecords
              RETURN "MaxRecords"

         CASE liIDProp=_DBCID_ShareConnection
              RETURN "ShareConnection"

         CASE liIDProp=_DBCID_SourceType
              RETURN "SourceType"

         CASE liIDProp=_DBCID_SQL
              RETURN "SQL"

         CASE liIDProp=_DBCID_Tables
              RETURN "Tables"

         CASE liIDProp=_DBCID_SendUpdates
              RETURN "SendUpdates"

         CASE liIDProp=_DBCID_Updatable
              RETURN "Updatable"

         CASE liIDProp=_DBCID_UpdateType
              RETURN "UpdateType"

         CASE liIDProp=_DBCID_UseMemoSize
              RETURN "UseMemoSize"

         CASE liIDProp=_DBCID_WhereType
              RETURN "WhereType"

         CASE liIDProp=_DBCID_DisplayClass
              RETURN "DisplayClass"

         CASE liIDProp=_DBCID_DisplayClassLibrary
              RETURN "DisplayClassLibrary"

         CASE liIDProp=_DBCID_Format
              RETURN "Format"

         CASE liIDProp=_DBCID_InputMask
              RETURN "InputMask"

         CASE liIDProp=_DBCID_Caption
              RETURN "Caption"

         CASE liIDProp=_DBCID_Asynchronous
              RETURN "Asynchronous"

         CASE liIDProp=_DBCID_BatchMode
              RETURN "BatchMode"

         CASE liIDProp=_DBCID_ConnectString
              RETURN "ConnectString"

         CASE liIDProp=_DBCID_ConnectTimeout
              RETURN "ConnectTimeout"

         CASE liIDProp=_DBCID_DispLogin
              RETURN "DispLogin"

         CASE liIDProp=_DBCID_DispWarnings
              RETURN "DispWarnings"

         CASE liIDProp=_DBCID_IdleTimeout
              RETURN "IdleTimeout"

         CASE liIDProp=_DBCID_QueryTimeout
              RETURN "QueryTimeout"

         CASE liIDProp=_DBCID_Password
              RETURN "Password"

         CASE liIDProp=_DBCID_Transactions
              RETURN "Transactions"

         CASE liIDProp=_DBCID_UserID
              RETURN "UserID"

         CASE liIDProp=_DBCID_WaitTime
              RETURN "WaitTime"

         CASE liIDProp=_DBCID_TimeStamp
              RETURN "TimeStamp"

         CASE liIDProp=_DBCID_DataType
              RETURN "DataType"

         CASE liIDProp=_DBCID_PacketSize
              RETURN "PacketSize"

         CASE liIDProp=_DBCID_Database
              RETURN "Database"

         CASE liIDProp=_DBCID_Prepared
              RETURN "Prepared"

         CASE liIDProp=_DBCID_CompareMemo
              RETURN "CompareMemo"

         CASE liIDProp=_DBCID_FetchAsNeeded
              RETURN "FetchAsNeeded"

         CASE liIDProp=_DBCID_Offline
              RETURN "Offline"

         CASE liIDProp=_DBCID_DBCEvents
              RETURN "DBCEvents"

         CASE liIDProp=_DBCID_DBCEventFileName
              RETURN "DBCEventFileName"

         CASE liIDProp=_DBCID_AllowSimultaneousFetch
              RETURN "AllowSimultaneousFetch"

         CASE liIDProp=_DBCID_DisconnectRollback
              RETURN "DisconnectRollback"

        OTHERWISE
             RETURN IIF(!ISNULL(This.oConnector),;
                    This.oConnector.FormatProperty(liIDProp),"")

      ENDCASE
      RETURN ""
   ENDPROC

   PROCEDURE CnToI(lcInt) && Convert 1-5 bytes string to Integer
      * lcInt - Binary format of integer
      *!*   <pdm_sc_yes/>
      LOCAL liLen,lii,liInt
      liLen=LEN(lcInt)
      DO CASE
         CASE liLen=1
              RETURN ASC(lcInt)
         CASE liLen=2
              RETURN This.C2ToI(lcInt)
         CASE liLen=4
              RETURN This.C4ToI(lcInt)

         OTHERWISE
              liInt=0
              FOR lii=1 TO LEN(lcInt)
                  liInt=liInt+ASC(SUBS(lcInt,lii,1))*2^((lii-1)*8)
              NEXT
              RETURN liInt
      ENDCASE

      RETURN 0
   ENDPROC

   PROCEDURE IToCn(liInt) && Convert Integer to 1-5 bytes string
      * liInt - Integer
      *!*   <pdm_sc_yes/>
      LOCAL lcVal,lii
      DO CASE
         CASE liInt<0x100
              RETURN CHR(liInt)
         CASE liInt<0x10000
              RETURN This.IToC2(liInt)
         CASE liInt<0x10000
              RETURN This.IToC4(liInt)

         OTHERWISE
             lcVal=""
             DO WHILE liInt>0
                 lii=INT(liInt/256) && Vydl 256
                 lcVal=lcVal+CHR(liInt-lii*256) && Peve rozdl na znak a piti ho
                 liInt=lii && Zapi novou hodnotu pro pevod
             ENDDO
      ENDCASE

      RETURN 0
   ENDPROC



   PROCEDURE C4ToI(lcInt)  && Convert binnary format to number (32)
      * lcInt - Binary format of integer
      *!*   <pdm_sc_yes/>
      RETURN ASC(LEFT(lcInt,1))+;
             ASC(SUBS(lcInt,2,1))*256+;
             ASC(SUBS(lcInt,3,1))*65536+;
             ASC(RIGHT(lcInt,1))*16777216
   ENDPROC


   PROCEDURE C4ToIRev(lcInt)  && Convert binnary format (reverz) to number (32)
      * lcInt - Binary format of integer
      *!*   <pdm_sc_yes/>
      RETURN ASC(LEFT(lcInt,1))*16777216+;
             ASC(SUBS(lcInt,2,1))*65536+;
             ASC(SUBS(lcInt,3,1))*256+;
             ASC(RIGHT(lcInt,1))
   ENDPROC


   PROCEDURE C2ToI(lcInt)  && Convert binnary format to number (16)
      * lcInt - Binary format of integer
      *!*   <pdm_sc_yes/>
      RETURN ASC(LEFT(lcInt,1))+ASC(RIGHT(lcInt,1))*256  
   ENDPROC


   PROCEDURE IToC2(liVal) && Convert number (16) to binnary format
      * liInt - Integer
      LOCAL lcVal,lii
      *!*   <pdm_sc_yes/>

      lii=INT(liVal/256) && Vydl 256
      lcVal=CHR(liVal-lii*256) && Peve rozdl na znak a piti ho
      liVal=lii && Zapi novou hodnotu pro pevod

      lii=INT(liVal/256) && Vydl 256
      RETURN lcVal+CHR(liVal-lii*256)
   ENDPROC


   PROCEDURE IToC4(liVal) && Convert number (32) to binnary format
      * liInt - Integer
      LOCAL lcVal,lii
      *!*   <pdm_sc_yes/>

      lii=INT(liVal/256) && Vydl 256
      lcVal=CHR(liVal-lii*256) && Peve rozdl na znak a piti ho
      liVal=lii && Zapi novou hodnotu pro pevod

      lii=INT(liVal/256) && Vydl 256
      lcVal=lcVal+CHR(liVal-lii*256)
      liVal=lii && Zapi novou hodnotu pro pevod

      lii=INT(liVal/256) && Vydl 256
      lcVal=lcVal+CHR(liVal-lii*256)
      liVal=lii && Zapi novou hodnotu pro pevod

      lii=INT(liVal/256) && Vydl 256
      RETURN lcVal+CHR(liVal-lii*256) && Peve rozdl na znak a piti ho
   ENDPROC

   PROCEDURE IToC4Rev(liVal) && Convert number (32) to binnary format (reverz)
      * liInt - Integer
      LOCAL lcVal,lii
      *!*   <pdm_sc_yes/>

      lii=INT(liVal/256) && Vydl 256
      lcVal=CHR(liVal-lii*256) && Peve rozdl na znak a piti ho
      liVal=lii && Zapi novou hodnotu pro pevod

      lii=INT(liVal/256) && Vydl 256
      lcVal=CHR(liVal-lii*256)+lcVal
      liVal=lii && Zapi novou hodnotu pro pevod

      lii=INT(liVal/256) && Vydl 256
      lcVal=CHR(liVal-lii*256)+lcVal
      liVal=lii && Zapi novou hodnotu pro pevod

      lii=INT(liVal/256) && Vydl 256
      RETURN CHR(liVal-lii*256)+lcVal && Peve rozdl na znak a piti ho
   ENDPROC


   PROCEDURE SetCatch(llCatch) && Catching error
      * llCatch - Mode cathing (.T. - catch, .F. - nocatch, .NULL. - nocatch,???? )
 
      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      LOCAL luCatch
      luCatch=This.Catch
      This.Catch=llCatch
      IF llCatch && If catching
         * the reset error and message number
         This.nErr=0 
         This.cErrMSG=""
      ENDIF
      RETURN luCatch && Return old catch flag
   ENDPROC


   PROCEDURE AttachSession(liSessionID) && Attach Session
      * liSessionID - Datasession ID

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      LOCAL luCatch
      luCatch=m.This.SetCatch(.T.)
      SET DATASESSION TO (liSessionID)
      =m.This.SetCatch(luCatch)
      RETURN m.This.nErr
   ENDPROC

   PROCEDURE OpenTable(lcFile,lcAlias,lcParam) && Open Table
      * lcFile  - Folder and name of table
      * lcAlias - Alias
      * lcParam - Parameters

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>
      LOCAL luCatch
      luCatch=m.This.SetCatch(.T.)
      USE (lcFile) ALIAS (lcAlias) IN 0 &lcParam.
      =m.This.SetCatch(luCatch)
      RETURN m.This.nErr
   ENDPROC

   PROCEDURE CloseTable(lcAlias) && Close table
      * lcAlias - Alias 
      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      IF USED(lcAlias)
         USE IN (lcAlias)
      ENDIF
      RETURN _DBC_ErrOK
   ENDPROC

   PROCEDURE Error(nError,cMethod,nLine) && Error routine
      * nError  - Error number
      * cMethod - Name of Method
      * nLine   - Number in method

      *!*   <pdm_sc_yes/>

      IF ISNULL(m.This.Catch)
         This.nErr=nError
         This.cErrMSG=MESSAGE() 
         ERROR nError
         RETURN
      ENDIF
      IF This.Catch && Catch
         This.nErr=nError
         This.cErrMSG=MESSAGE()
      ELSE
         ERROR nError
      ENDIF
   ENDPROC



ENDDEFINE

*********************************************************************************
*
*********************************************************************************
DEFINE CLASS _DBC_Check AS _DBC && Class for checking DBC
   Name="_DBC_Check"

   PROCEDURE Check(lcAlias) && Check DBC integrity
      * lcAlias - Alias of opened DBC

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      LOCAL loInfo
      loInfo=CREATEOBJECT("_DBC_CheckInfo")

      =This.FindLostObjects(lcAlias,loInfo)
      =This.FindObsoleteObjects(lcAlias,loInfo)
      =This.CheckPropertiesOfAllObjects(lcAlias,loInfo)
      =This.CheckSPFXP(lcAlias,loInfo)

      =This.CreateLog(loInfo)
   ENDPROC


   PROCEDURE FindLostObjects(lcAlias,loInfo) && Find lost objects
      * lcAlias - Alias of opened DBC
      * loInfo  - Info object

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>
      SELECT AA.OBJECTID FROM (lcAlias) AA;
             WHERE NOT DELETED() AND NOT AA.PARENTID IN (SELECT AB.OBJECTID FROM (lcAlias) AB);
             INTO ARRAY loInfo.aLostObjects

      loInfo.iLostObjects=_TALLY
      RETURN 
   ENDPROC

   PROCEDURE FindObsoleteObjects(lcAlias,loInfo) && Find obsolete objects
      * lcAlias - Alias of opened DBC
      * loInfo  - Info object

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>
      SELECT AA.OBJECTID FROM (lcAlias) AA;
             WHERE DELETED();
             INTO ARRAY loInfo.aObsoleteObjects

      loInfo.iObsoleteObjects=_TALLY
      RETURN 
   ENDPROC

   PROCEDURE CheckPropertiesOfAllObjects(lcAlias,loInfo) && Check properties of all objects
      * lcAlias    - Alias of opened DBC
      * loInfo - Info object

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      LOCAL lii,liCount,liy
      LOCAL ARRAY laProp(1)

      liy=0
      SELECT (lcAlias)
      SCAN FOR !DELETED() AND OBJECTNAME#PADR("TransactionLog",128) AND ;
               OBJECTNAME#PADR("StoredProceduresSource",128) AND ;
               OBJECTNAME#PADR("StoredProceduresObject",128)

           liCount=This.CheckProperties(lcAlias,OBJECTID,@laProp) && Check Property's value

           FOR lii=1 TO liCount
               IF laProp(lii,4)#_DBC_Check_OK
                  liy=liy+1
                  DIMENSION loInfo.aInvalidProperties(liy)
                  loInfo.aInvalidProperties(liy)=CREATEOBJECT("_DBC_CheckInfo_InvalidProperties")
                  loInfo.aInvalidProperties(liy).ObjectID=OBJECTID
                  ACOPY(laProp,loInfo.aInvalidProperties(liy).aInvalidProperties)
                  loInfo.aInvalidProperties(liy).iInvalidProperties=liCount
                  EXIT
               ENDIF
           NEXT

      ENDSCAN
      loInfo.iInvalidProperties=liy
   ENDPROC


   PROCEDURE CheckSPFXP(lcAlias,loInfo) && Check stored procedures FXP code
      * lcAlias - Alias of opened DBC
      * loInfo  - Info object

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      * backup SET DEFA
      * set DBC path as SET DEFA
      * save SP as PRG and recompile it
      * Get SP FXP and compare it with recompiled source code

      LOCAL lcOldDEFA,lcSC,lcFXP,lcTMP,lcFRXPN,luCatch
      lcOldDEFA=SET("DEFA")+CURDIR()

      SET DEFAULT TO (JUSTPATH(DBF(lcAlias))+"\")
      * Get Stored Procedure
      =loDBC.GetSP(lcAlias,_DBC_SP_TSource,@lcSC)
      =loDBC.GetSP(lcAlias,_DBC_SP_TFXP,@lcFXP)
      lcTMP=SYS(2023)+"\"+SYS(2015)+".prg"
      loInfo.SPFXP=_DBC_SP_CheckFXP_OK

      luCatch=m.This.SetCatch(.T.)
      =STRTOFILE(lcSC,lcTMP)
      IF This.nErr#0
         loInfo.SPFXP=_DBC_SP_CheckFXP_CantSavePRG
      ENDIF
      IF This.nErr=0
         =m.This.SetCatch(.T.)
         COMPILE (lcTMP) 
         IF This.nErr#0
            loInfo.SPFXP=_DBC_SP_CheckFXP_CantCompilePRG
         ENDIF
      ENDIF     

      IF This.nErr=0
         =m.This.SetCatch(.T.)
         lcFXPN=FILETOSTR(FORCEEXT(lcTMP,"fxp"))
         IF This.nErr#0
            loInfo.SPFXP=_DBC_SP_CheckFXP_CantReadFXP
         ENDIF
      ENDIF     

      IF This.nErr=0
         loInfo.SPFXP = IIF(lcFXP==lcFXPN,_DBC_SP_CheckFXP_OK,_DBC_SP_CheckFXP_Failed) && Stored procedures FXP status
         IF FILE(FORCEEXT(lcTMP,"err"))
            =m.This.SetCatch(.T.)
            loInfo.SPFXPResult = FILETOSTR(FORCEEXT(lcTMP,"err")) && Stored procedures Error result
            IF This.nErr#0
               loInfo.SPFXP=_DBC_SP_CheckFXP_CantReadErr
            ENDIF
         ENDIF
      ENDIF

      DELETE FILE (lcTMP)
      DELETE FILE (FORCEEXT(lcTMP,"fxp"))
      DELETE FILE (FORCEEXT(lcTMP,"err"))
      SET DEFAULT TO (lcOldDEFA)
      =m.This.SetCatch(luCatch)
   ENDPROC


   PROCEDURE CheckTablesStructure(lcAlias,loInfo) && Check tables structure
      * lcAlias - Alias of opened DBC
      * loInfo  - Info object

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

   ENDPROC

   PROCEDURE CreateLog(loInfo) && Create output file
      * loInfo - Info object

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

** invalid properties
*!*              IF lii<=liCount
*!*                 ?
*!*                 ?ALLTRIM(OBJECTNAME)
*!*                 FOR lii=1 TO liCount
*!*                     IF laProp(lii,4)#_DBC_Check_OK
*!*                        ?DBC_FormatProperty(laProp(lii,1)),laProp(lii,3),laProp(lii,4)
*!*                     ENDIF
*!*                 NEXT
*!*             ENDIF

   ENDPROC

ENDDEFINE

*********************************************************************************
*
*********************************************************************************
DEFINE CLASS _DBC_CheckInfo AS custom && Class about checking DBC
   Name="_DBC_CheckInfo"

   iLostObjects=0 && Count of Lost objects
   DIME aLostObjects(1) && List of Lost objects

   iObsoleteObjects=0 && Count of obsolete objects
   DIME aObsoleteObjects(1) && List of obsolete objects

   iInvalidProperties=0 && Count of objects which have invalid properties
   DIME aInvalidProperties(1) && List of objects which have invalid properties

   SPFXP = _DBC_SP_CheckFXP_OK && Stored procedures FXP status
   SPFXPEx = _DBC_SP_CheckFXP_OK && Extended Stored procedures status
   SPFXPResult = "" && Stored procedures Error result
ENDDEFINE

*********************************************************************************
*
*********************************************************************************
DEFINE CLASS _DBC_CheckInfo_InvalidProperties AS custom && Class about checking DBC - invalid properties
   Name="_DBC_CheckInfo_InvalidProperties"

   ObjectID=0 && Object ID
   iInvalidProperties=0 && Count of objects which have invalid properties
   DIME aInvalidProperties(1) && List of objects which have invalid properties

ENDDEFINE

*********************************************************************************
*
*********************************************************************************
DEFINE CLASS _DBC_Connector AS custom && DBC connector for user custom properties
   Name="_DBC_Connector"

   PROCEDURE FormatProperty(liIDProp) && Convert property ID to Name
      * liIDProp   - Property ID

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      RETURN "" && default return value
   ENDPROC


   PROCEDURE GetDefaultValue(liIDProp) && Get default Value
      * liIDProp   - Property ID

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      RETURN .NULL. && default return value
   ENDPROC


   PROCEDURE GetValueType(liIDProp) && Return the property type
      * liIDProp - Property ID

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      RETURN _DBC_ValueType_Unknown && default return value
   ENDPROC


   PROCEDURE DBCValueToValue(liType,lcValue) && Convert DBC value to Value
      * liType  - Property type
      * luValue - Property value

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      RETURN lcValue
   ENDPROC


   PROCEDURE ValueToDBCValue(liType,luValue) && Convert value to DBC value
      * liType  - Property type
      * luValue - Property value

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      RETURN .NULL.
   ENDPROC


   PROCEDURE CheckValue() && Check Value
      LPARAMETERS liIDProp,liType,luValue
      * liIDProp   - Property ID
      * liType     - Value type
      * luValue    - Value

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      * 1) check type
      * 2) check interval
      * 4) return .T. is value valid or .F. is invalid
   
      RETURN .T.
   ENDPROC


   PROCEDURE CheckProperties() && Check property value (DBC value)
      LPARAMETERS liIDProp,liType,lcValue
      * liIDProp   - Property ID
      * liType     - Value type
      * lcValue    - DBC value

      *!*   <pdm_sc_yes/>
      *!*   <pdm_dd_yes/>

      RETURN .T.
   ENDPROC

ENDDEFINE
